//car.cpp - This file is meant to be left as is, except for the device settings 

////////////////////////////////////////////////////////////////////////////////
//  car.cpp and car.h contain all core functions for webpage, WiFi, driving
//      The user does not need to change this code
//      Uses string definitions from webPage.cpp as exposed in webPage.h 
//      Calls pre-defined user functons in functions.cpp, exposed in functions.h

#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <ESP32Servo.h>  // older version <ESP32_Servo.h>, newer version same library <ESP32Servo.h>
#include <WString.h>
#include "DNSServer.h"

#include "car.h"
#include "webPage.h"
#include "functions.h"

namespace car{
  //////////////////////////////////////////////////////////////////////////////
  /////////////// Device Settings - which the user can change here /////////////

  //////////////////////////////////////////////////////////////////////////////
  // robot WiFi access point settings
  //   optional is the use of an existing WiFi network  
  //      which allows you to use internet and controll the robot at same time 
  //   if not using an existing WiFi network, robot exposes a WiFi access point
  //      with a captive portal to easier access the robot controll page
  //////////// WiFi and networking settings ////////////////////////////////////

  // settings for joining an existing WiFi network
  const char* ssid_Client     = "Junior IOT RobotBattle";
  const char* password_Client = "langebus";
  bool connectToWiFi = true; // if WiFi network not found, robot becomes sluggish
  // If connectToWiFi is set to true, no captive portal is created

  // settings for robot as an access point
  String ssid_Host     = "DISK";
  bool concatinate_mac_address_to_ssid_Host = true;
  const char* password_Host = "passwordsecret";
  bool passwordProtect = false;

  // setting to help to stop the robot when nobody is controlling
  // the next variable is used in an interrupt, so it should be volatile
  volatile int timeout_command_received = 5000; // milliseconds, larger than 100

  
  //////////////////////////////////////////////////////////////////////////////
  //////////////// Defines - do not change - common    /////////////////////////
  bool logData = true;
  
 
  //                         L           L            R           R
  int motorPins[4] = {MOTOR_PIN_L1, MOTOR_PIN_L2, MOTOR_PIN_R1, MOTOR_PIN_R2};

  // non-tank drive speed and direction received from controll page
  int throttle = 0;
  int rot     = 0;

  // declarations
  Adafruit_NeoPixel pixels(NEOPIXEL_NUMPIXELS, NEOPIXEL_PIN, NEO_GRB + NEO_KHZ800);
  void neopixel_ipaddress_update();
  IPAddress NeoPixel_IP_display;
  int NeoPixel_display_local_wifi;
  const char* wl_status_to_string(wl_status_t status);

  //////////////// Defines - do not change - Networking    ////////////////////
  // the next variable is used in an interrupt, so it should be volatile
  volatile long time_since_last_websockets_data = 0;  
  hw_timer_t* timer = NULL;  // timer for interrupting
  
  AsyncWebServer server(80);
  String server_join_message = "<h1>Please connect to http://www.junioriot.nl<\h1>";
  AsyncWebSocket ws("/ws");
  
  DNSServer dnsServer;  // creates DNS server, required for captive portal

  TaskHandle_t carTask;
  
  //////////////////////////////////////////////////////////////////////////////
  ///////////////  init any functions for car operation           //////////////
  
  void initCar(){        
    // init the motors by attaching the pins
    initMotor();
    
    // concatinate the mac address to the ssid if settings are set for it
    if(concatinate_mac_address_to_ssid_Host){
      ssid_Host = String(ssid_Host) + String("-")
           + String((uint32_t)ESP.getEfuseMac(), HEX);
    }

    // setup serial for readable output
    Serial.begin(115200);
    Serial.println();    
    Serial.println("Starting robot '" + String(ssid_Host) + "'");

    // INITIALIZE NeoPixel strip object (REQUIRED)
    pixels.begin(); 

    // setup timeout interrupt to detect if we lost connection from web controll page
    timer = timerBegin(0, 80, true);
    timerAttachInterrupt(timer, &checkTimeout, true);
    timerAlarmWrite(timer, 100000, true); //run function every 100 ms
    timerAlarmEnable(timer);

    // create wifi access point host on the robot
    if(passwordProtect){
      WiFi.softAP(ssid_Host.c_str(), password_Host);
    }else{
      WiFi.softAP(ssid_Host.c_str());
    }

    // if requested, connect to wifi network as client
    if(connectToWiFi) {
      WiFi.begin(ssid_Client, password_Client);
    }
    // we need to allow some time to receive a client IP number
    
    // configure the webserver (websockets and webpages)
    ws.onEvent(onWebsocket);
    server.addHandler(&ws);
    
    server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
      request->send(200, "text/html", webPage::pageC);
    });

    // create the web response for the captive portal
    //   captive displays a message instead of redirecting to the main page 
    //     because andriod's built in browser is not compatible with the webpage   
    // User needs to go to any page on Android, or to robot ip address on iPhone 
    server.on("/generate_204", HTTP_GET, [](AsyncWebServerRequest *request){      
      // next variable needs to be declared in same scope as 'server'
      //     or the next line may generate an error: variable is not captured
      server_join_message = "<h1>Please connect to http://www.junioriot.nl<\h1> <br> or http://" +
              String(WiFi.softAPIP()) + " or http://" + String(NeoPixel_IP_display);
      request->send(200, "text/html", server_join_message );
      Serial.println("requested /generate_204");
    });

    // redirect all unknown web/dns requests to robot controll page
    server.onNotFound([](AsyncWebServerRequest *request){
      request->redirect("/");
    });

    // Setup dns for captive login page
    if(connectToWiFi == false){
      xTaskCreate(
                      dnsTask,          // Task function. 
                      "dnsTask",        // String with name of task. 
                      10000,            // Stack size in bytes. 
                      NULL,             // Parameter passed as input of the task 
                      1,                // Priority of the task. 
                      NULL);            // Task handle. 
    }

    // start the webserver
    server.begin();

    // now allow some time to receive stable WiFi status;  
    long start_time=millis();  
    Serial.println("WiFi Status: " + String(wl_status_to_string(WiFi.status())));
    if (String(wl_status_to_string(WiFi.status())) !="WL_CONNECTED") {
      Serial.print("Wait for WiFi connection or timeout");
      while((String(wl_status_to_string(WiFi.status())) !="WL_CONNECTED") && (millis()-start_time)< 5000) {
        // wait a bit more until status is good
        delay(100);
        Serial.print(".");
      }
      Serial.println("\nWiFi Status is now: " + String(wl_status_to_string(WiFi.status())));
    }

    // get the IP´s of the device
    IPAddress HostIP = WiFi.softAPIP();
    IPAddress ClientIP = WiFi.localIP();
    NeoPixel_IP_display = HostIP;
    NeoPixel_display_local_wifi=0;
    Serial.println("Robot hosted WiFi is '" + String(ssid_Host) + "'");
    Serial.print("  with the robot controll page at 'http://");
    Serial.print(HostIP);
    Serial.println("'");
    if(connectToWiFi && ClientIP[3] > 0){
      NeoPixel_IP_display = ClientIP;
      NeoPixel_display_local_wifi=1;
      Serial.println("Robot joined a local WiFi network '" + String(ssid_Client) + "'");
      Serial.print("  with the robot controll page at: 'http://");
      Serial.print(ClientIP);
      Serial.println("'");
    } else if(connectToWiFi && !(ClientIP[3] >0)){
      Serial.println("Robot failed to join local WiFi network '" + String(ssid_Client) + "'");
    }else{
      Serial.println("Robot was configured not to join a local WiFi network.");
    }

    ///////////////////////////////////////////////////////////////////////////
    //  run setup function once to start any custom items from functions.cpp
    setupFunction();

    ///////////////////////////////////////////////////////////////////////////
    //  repeat loopFunction to run any custom items from functions.cpp
    
    while(true){
      // some housekeeping core functions 
      neopixel_ipaddress_update();

      // run any custom items from functions.cpp
      loopFunction();
    }

    // Note: this line will not be reached due to repeating loopFunction
    ///////////////////////////////////////////////////////////////////////////
  }



  //////////////////////////////////////////////////////////////////////////////
  ///////////////  Required core functions for NeoPixel

  // display the ip address on the NeoPixel strip
  bool ipaddress_is_displaying=false;
  int prev_step_ipaddress=-1;
  unsigned long ipaddress_display_started = -1;
    
  bool neopixel_is_available() {  
    bool is_blocked = ipaddress_is_displaying;    
    return !is_blocked;
  }

  void neopixel_ipaddress_start() {  
    // main loop runs very frequently --> use millis() to calculate status
    unsigned long this_moment = millis();
    
    // init 
    if(neopixel_is_available()) {
      // start displaying the IP address
      ipaddress_is_displaying=true;
      ipaddress_display_started=this_moment;
      //Serial.print("NeoPixel IP... ");
    }    
  }

  void neopixel_ipaddress_update() {
    // main loop runs very frequently --> use millis() to calculate status
    unsigned long this_moment = millis();
    
    // update the display if needed
    const int time_per_byte = 3000; // milisec
    const int time_pause = 500; // milisec
    //const int time_total = time_per_byte * 4 + time_pause * 5; // 14 steps
    int curr_time = this_moment - ipaddress_display_started;
    int curr_step = -1;
    int r, g, b, p; // p=power
    int my_byte;

    if (neopixel_is_available()){
      // not displaying, so do nothing
    } else {
      // find out which display step is needed (ugly hardcoded approach)
      if      (curr_time < (1*time_pause + 0*time_per_byte)){  // dark
        curr_step = 1; p=0;
      } else if (curr_time < (1*time_pause + 1*time_per_byte)){ // starter flash
        curr_step = 2;   
      } else if (curr_time < (2*time_pause + 1*time_per_byte)){  // dark
        curr_step = 3; p=0;
      } else if (curr_time < (2*time_pause + 2*time_per_byte)){  // first byte
        curr_step = 4; p=1; r=255; g=0; b=0; my_byte=NeoPixel_IP_display[0];  
      } else if (curr_time < (3*time_pause + 2*time_per_byte)){  // dark
        curr_step = 5; p=0;
      } else if (curr_time < (3*time_pause + 3*time_per_byte)){  // second byte
        curr_step = 6; p=1; r=0; g=255; b=0; my_byte=NeoPixel_IP_display[1];  
      } else if (curr_time < (4*time_pause + 3*time_per_byte)){  // dark
        curr_step = 7; p=0;
      } else if (curr_time < (4*time_pause + 4*time_per_byte)){  // third byte
        curr_step = 8; p=1; r=0; g=0; b=255; my_byte=NeoPixel_IP_display[2]; 
      } else if (curr_time < (5*time_pause + 4*time_per_byte)){  // dark
        curr_step = 9; p=0;
      } else if (curr_time < (5*time_pause + 5*time_per_byte)){  // fourth byte 
        curr_step =10; p=1; r=250; g=202; b=48; my_byte=NeoPixel_IP_display[3]; 
      } else if (curr_time < (8*time_pause + 5*time_per_byte)){ // dark
        curr_step =11; p=0;
      } else {
        // after more time, send a signal that NeoPixel is now available
        ipaddress_is_displaying=false;
        //Serial.println("");
      }
  
      if (ipaddress_is_displaying && (curr_step != prev_step_ipaddress) ) {
        // we need to display something new
        pixels.clear(); // Set all pixel colors to 'off'
        if (curr_step==2) {
          // display starter flash  (ugly hardcoded approach)
          pixels.setPixelColor(0, pixels.Color(255, 0, 0));
          pixels.setPixelColor(1, pixels.Color(255, 0, 0));
          pixels.setPixelColor(2, pixels.Color(0, 255, 0));
          pixels.setPixelColor(3, pixels.Color(0, 255, 0));
          pixels.setPixelColor(4, pixels.Color(0, 0, 255));
          pixels.setPixelColor(5, pixels.Color(0, 0, 255));
          if(NeoPixel_display_local_wifi) {
            pixels.setPixelColor(6, pixels.Color(255, 255, 255));
            pixels.setPixelColor(7, pixels.Color(255, 255, 255));
          }
        } else if (p==0) {
          // display dark
        } else {
          // display a byte from IP address
          //Serial.print(" " + String(my_byte) + "=");
          for(int i=0;i<8;i++) { //go up from led 0 to 7
            if( ((int) my_byte & (int) pow(2,i) )>0 ) {  // & is the bitwise and operator
              // pixels.Color() takes RGB values, from 0,0,0 up to 255,255,255
              pixels.setPixelColor(i, pixels.Color(r, g, b));
              //Serial.print("1");
            }else{
              //Serial.print("0");
            }
          }   
          //Serial.print(" in rgb(" + String(r) +"," + String(g) +"," + String(b) +") ");
        }
        
        pixels.show();   // Send the updated pixel colors to the hardware.
        prev_step_ipaddress = curr_step;
      }  // if a change was needed
    } // if we are displaying ip address
  }
  
  //////////////////////////////////////////////////////////////////////////////
  ///////////////  Required core functions for Networking and connectivity
  
  // actions for when the robots disconnects from controll page (watchdog driven)
  void disconnected_actions() {  
    //Stop the robot when last client is disconnected
    movement(0, 0); 
    // when reconnected, calls from web page will start movement immediately

    // display IP address on NeoPixels so user can find web page
    neopixel_ipaddress_start();  
    // when reconnected, ip display cycle will still complete
  }

  //websocket handle function (event driven actions)
  void onWebsocket(AsyncWebSocket * server, AsyncWebSocketClient * client, AwsEventType type, void * arg, uint8_t *data, size_t len){
    AwsFrameInfo * info = (AwsFrameInfo*)arg;

    if(type == WS_EVT_CONNECT){
      Serial.println("Client '" + String(client->id()) + "' has connected");
    } else if(type == WS_EVT_PONG){
      Serial.println("Client '" + String(client->id()) + "' has answered ping request");
    } else if(type == WS_EVT_DISCONNECT){
      Serial.println("Client '" + String(client->id()) + "' has been disconnected");
      disconnected_actions();
    } else if(type == WS_EVT_ERROR){
      Serial.println("Client '" + String(client->id()) + "': an error occured");
      disconnected_actions();
    } else if(type == WS_EVT_DATA){
      if(info->opcode == WS_TEXT){
        //text was received from javascript code in our controll webpage

        //reset the timeout timer
        time_since_last_websockets_data = millis();
        
        //find the command sent by the js code
        int key = (char)data[0];
        
        //build the value string out of received chars
        String value;
        for(size_t i=1; i < info->len; i++){
          value += (char)data[i];
        }

        //run the code which handles the commands sent by the js code
        switch(key){
          case 't':
            throttle = value.toInt();
            break;
          case 'r':
            rot = value.toInt();
            break;
          case 'b':
            if(value.toInt() == 1){
              button1();
            }else if(value.toInt() == 2){
              button2();
            }else if(value.toInt() == 3){
              button3();
            }else if(value.toInt() == 4){
              button4();
            }
            break;
        }
        movement(throttle, rot);
      }else{
        //binary message was received, do nothing, our js code shouldn´t send binary messages
      }
    }
  }

  //check timeoutwatchdog, as an interrupt
  void IRAM_ATTR checkTimeout(){                                          
    if(millis() - time_since_last_websockets_data > timeout_command_received){
      // when we see no recent activity 
      disconnected_actions();      
    }
  }

  //task for dns server handle, redirects to controll page as part of captive portal
  void dnsTask( void * parameter ){
    // start the DNS server, "*" sends request for port 53 to IP address of the device
    dnsServer.start(53, "*", WiFi.softAPIP());  
    
    while(true){
      dnsServer.processNextRequest(); // Without this, the connected device will simply timeout trying to reach the internet
                                     // or it might fall back to mobile data if it has that
                                     
      delay(100); //free up the core in this repeated task
    }
  }

  const char* wl_status_to_string(wl_status_t status) { // wifi status 
    switch (status) {
      case WL_NO_SHIELD: return "WL_NO_SHIELD";
      case WL_IDLE_STATUS: return "WL_IDLE_STATUS";
      case WL_NO_SSID_AVAIL: return "WL_NO_SSID_AVAIL";
      case WL_SCAN_COMPLETED: return "WL_SCAN_COMPLETED";
      case WL_CONNECTED: return "WL_CONNECTED";
      case WL_CONNECT_FAILED: return "WL_CONNECT_FAILED";
      case WL_CONNECTION_LOST: return "WL_CONNECTION_LOST";
      case WL_DISCONNECTED: return "WL_DISCONNECTED";
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  ///////////////  Required core functions for Motors
  
  // init the motors by attaching pins to the pwm device (internal in esp)
  void initMotor(){
    for(int i = 1; i <= 4; i++){
      ledcSetup(i+4, 5000, 8);  //use channel 5, 6, 7, 8
      ledcAttachPin(motorPins[i - 1], i+4);
    }
  }

  // convert integer into motor compatable value (according to min and max values)
  int toMotorSpeed(int x){
    if(x < -255) x = -255;
    if(x > 255) x = 255;
    return x;
  }

  // return the Throttle defined by webpage
  int getThrottle(){
    return throttle;
  }

  // return the rotation defined by webpage
  int getRotation(){
    return rot;
  }

  // write motor Throttle value to motor
  void writeMotorL(int motorThrottle){
    //make sure Throttle value is compatable with motor
    motorThrottle = toMotorSpeed(motorThrottle);

    //find movement rotation and write the speed to the correct motor pin
    if(motorThrottle > 0){
      ledcWrite(1+4, motorThrottle);
      ledcWrite(2+4, 0);
    }else{
      ledcWrite(1+4, 0);
      ledcWrite(2+4, -motorThrottle); //  -motorThrottle is positive
    }
  }

  // write motor Throttle value to motor
  void writeMotorR(int motorThrottle){
    //make sure Throttle value is compatable with motor
    motorThrottle = toMotorSpeed(motorThrottle);

    //find movement rotation and write the speed to the correct motor pin
    if(motorThrottle > 0){
      ledcWrite(3+4, motorThrottle);
      ledcWrite(4+4, 0);
    }else{
      ledcWrite(3+4, 0);
      ledcWrite(4+4, -motorThrottle);    
    }
  }

  // write both motors in the same function call
  void writeMotors(int motorL, int motorR){
    writeMotorL(motorL);
    writeMotorR(motorR);
  }


  /////////////////////////////////////////////////////////////////////////////
}
